@@ -244,7 +244,11 @@ impl<E> CorsEndpoint<E> {
         )
     }
 
-    fn build_preflight_response(&self, origin: &HeaderValue) -> Response {
+    fn build_preflight_response(
+        &self,
+        origin: &HeaderValue,
+        request_headers: Option<&HeaderValue>,
+    ) -> Response {
         let mut builder = Response::builder()
             .header(header::ACCESS_CONTROL_ALLOW_ORIGIN, origin)
             .typed_header(self.expose_headers_header.clone())
@@ -272,7 +276,11 @@ impl<E> CorsEndpoint<E> {
         }
 
         if self.allow_headers.is_empty() {
-            builder = builder.header(header::ACCESS_CONTROL_ALLOW_HEADERS, "*");
+            if let Some(request_headers) = request_headers {
+                builder = builder.header(header::ACCESS_CONTROL_ALLOW_HEADERS, request_headers);
+            } else {
+                builder = builder.header(header::ACCESS_CONTROL_ALLOW_HEADERS, "*");
+            }
         } else {
             builder = builder.typed_header(self.allow_headers_header.clone());
         }
@@ -283,6 +291,33 @@ impl<E> CorsEndpoint<E> {
 
         builder.body(())
     }
+
+    fn check_allow_headers<'a>(&self, req: &'a Request) -> (bool, Option<&'a HeaderValue>) {
+        let mut allow_headers = true;
+
+        let request_headers = if let Some(request_header) =
+            req.headers().get(header::ACCESS_CONTROL_REQUEST_HEADERS)
+        {
+            if !self.allow_headers.is_empty() {
+                allow_headers = false;
+                if let Ok(s) = request_header.to_str() {
+                    for header in s.split(',') {
+                        if let Ok(header) = HeaderName::from_str(header.trim()) {
+                            if self.allow_headers.contains(&header) {
+                                allow_headers = true;
+                                break;
+                            }
+                        }
+                    }
+                }
+            }
+            Some(request_header)
+        } else {
+            None
+        };
+
+        (allow_headers, request_headers)
+    }
 }
 
 #[async_trait::async_trait]
@@ -320,33 +355,13 @@ impl<E: Endpoint> Endpoint for CorsEndpoint<E> {
                 return Err(Error::new(StatusCode::UNAUTHORIZED));
             }
 
-            let allow_headers = {
-                let mut allow_headers = true;
-                if !self.allow_headers.is_empty() {
-                    if let Some(request_header) =
-                        req.headers().get(header::ACCESS_CONTROL_REQUEST_HEADERS)
-                    {
-                        allow_headers = false;
-                        if let Ok(s) = request_header.to_str() {
-                            for header in s.split(',') {
-                                if let Ok(header) = HeaderName::from_str(header.trim()) {
-                                    if self.allow_headers.contains(&header) {
-                                        allow_headers = true;
-                                        break;
-                                    }
-                                }
-                            }
-                        }
-                    }
-                }
-                allow_headers
-            };
+            let (allow_headers, request_headers) = self.check_allow_headers(&req);
 
             if !allow_headers {
                 return Err(Error::new(StatusCode::UNAUTHORIZED));
             }
 
-            return Ok(self.build_preflight_response(&origin));
+            return Ok(self.build_preflight_response(&origin, request_headers));
         }
 
         let mut resp = self.inner.call(req).await.into_response();
@@ -498,7 +513,7 @@ mod tests {
             resp.headers()
                 .get(header::ACCESS_CONTROL_ALLOW_HEADERS)
                 .unwrap(),
-            "*"
+            "X-Token"
         );
         assert_eq!(
             resp.headers().get(header::ACCESS_CONTROL_MAX_AGE).unwrap(),
@@ -742,4 +757,31 @@ mod tests {
             .get(header::ACCESS_CONTROL_ALLOW_ORIGIN)
             .is_none());
     }
+
+    #[tokio::test]
+    async fn allow_all_access_control_allow_headers_should_return_with_request_headers() {
+        let ep = make_sync(|_| "hello").with(
+            Cors::new()
+                .allow_origin(ALLOW_ORIGIN)
+                .allow_method(Method::GET),
+        );
+        let resp = ep
+            .map_to_response()
+            .call(
+                Request::builder()
+                    .method(Method::OPTIONS)
+                    .header(header::ORIGIN, ALLOW_ORIGIN)
+                    .header(header::ACCESS_CONTROL_REQUEST_METHOD, "GET")
+                    .header(header::ACCESS_CONTROL_REQUEST_HEADERS, "content-type")
+                    .finish(),
+            )
+            .await;
+        assert_eq!(resp.status(), StatusCode::OK);
+        assert_eq!(
+            resp.headers()
+                .get(header::ACCESS_CONTROL_ALLOW_HEADERS)
+                .unwrap(),
+            "content-type"
+        )
+    }
 }
